"""Greedy coloring test suite."""
import pytest
import networkx as nx
from collections import defaultdict
import gcol


GREEDY_METHODS = ["random", "welsh_powell", "dsatur", "rlf"]
OPT_ALGS = [1, 2, 3, None]
IT_LIMITS = [0, 100, 1000]
VERBOSE = [0, 1, 2]


class TestColorings:
    def test_many(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            for strategy in GREEDY_METHODS:
                for opt_alg in OPT_ALGS:
                    for it_limit in IT_LIMITS:
                        for verbose in VERBOSE:
                            c = gcol.node_coloring(
                                G,
                                strategy=strategy,
                                opt_alg=opt_alg,
                                it_limit=it_limit,
                                verbose=verbose
                            )
                            assert verify_node_coloring(G, c)
                            c = gcol.node_colouring(
                                G,
                                strategy=strategy,
                                opt_alg=opt_alg,
                                it_limit=it_limit,
                                verbose=verbose
                            )
                            assert verify_node_coloring(G, c)
                            c = gcol.edge_coloring(
                                G,
                                strategy=strategy,
                                opt_alg=opt_alg,
                                it_limit=it_limit,
                                verbose=verbose
                            )
                            assert verify_edge_coloring(G, c)
                            c = gcol.edge_colouring(
                                G,
                                strategy=strategy,
                                opt_alg=opt_alg,
                                it_limit=it_limit,
                                verbose=verbose
                            )
                            assert verify_edge_coloring(G, c)

    def test_bad_strategy(self):
        graph = singleton()
        pytest.raises(
            ValueError,
            gcol.node_coloring,
            graph,
            strategy="this is an invalid strategy",
        )

    def test_bad_opt_alg(self):
        graph = singleton()
        pytest.raises(
            ValueError,
            gcol.node_coloring,
            graph,
            opt_alg="this is an invalid optimisation algorithm",
        )

    def test_bad_its_parameter(self):
        graph = singleton()
        pytest.raises(
            ValueError,
            gcol.node_coloring,
            graph,
            it_limit="this is not an integer",
        )

    def test_negative_tabu_parameter(self):
        graph = singleton()
        pytest.raises(ValueError, gcol.node_coloring, graph, it_limit=-1)

    def test_bad_verbose_parameter(self):
        graph = singleton()
        pytest.raises(
            ValueError,
            gcol.node_coloring,
            graph,
            opt_alg=1,
            verbose="this is not an integer"
        )

    def test_negative_verbose_parameter(self):
        graph = singleton()
        pytest.raises(ValueError, gcol.node_coloring, graph, verbose=-1)

    def test_directed_graph(self):
        graph = nx.erdos_renyi_graph(10, 0.5, directed=True)
        pytest.raises(NotImplementedError, gcol.node_coloring, graph)

    def test_multigraph(self):
        graph = nx.MultiGraph()
        graph.add_edges_from(
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, 0), (0, 1), (0, 1)])
        pytest.raises(NotImplementedError, gcol.node_coloring, graph)

    def test_loops(self):
        graph = nx.complete_graph(3)
        graph.add_edge(0, 0)
        pytest.raises(NotImplementedError, gcol.node_coloring, graph)


class TestChromatics:
    def test_many_chromatic_number(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            chi = gcol.chromatic_number(G)
            assert chi >= 0 and chi <= G.number_of_nodes()

    def test_many_chromatic_index(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            chi = gcol.chromatic_index(G)
            delta = get_max_degree(G)
            assert chi == delta or chi == delta + 1


class TestNodePrecolorings:
    def test_many(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            # color all nodes, then uncolor some to get a valid precoloring
            precol = gcol.node_coloring(G, strategy="random")
            colSize = defaultdict(int)
            for u in precol:
                colSize[precol[u]] += 1
            V = list(precol)
            for i in range(len(V)):
                u = V[i]
                if i % 2 == 0 and colSize[precol[u]] > 1:
                    colSize[precol[u]] -= 1
                    del precol[u]
            for strategy in GREEDY_METHODS:
                for opt_alg in OPT_ALGS:
                    for it_limit in IT_LIMITS:
                        c = gcol.node_precoloring(
                            G,
                            precol,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit,
                        )
                        assert verify_node_coloring(G, c, precol)
                        c = gcol.node_precolouring(
                            G,
                            precol,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit,
                        )
                        assert verify_node_coloring(G, c, precol)
                        c = gcol.node_precoloring(
                            G,
                            None,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit,
                        )
                        assert verify_node_coloring(G, c)
                        c = gcol.node_precoloring(
                            G,
                            {},
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit
                        )
                        assert verify_node_coloring(G, c, {})

    def test_bad_precol1(self):
        graph = singleton()
        pytest.raises(TypeError, gcol.node_precoloring,
                      graph, precol="this is not a dict")

    def test_bad_precol2(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.node_precoloring, graph, precol={4: 1})

    def test_bad_precol3(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.node_precoloring,
                      graph, precol={0: "color 5"})

    def test_bad_precol4(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.node_precoloring,
                      graph, precol={0: 0, 1: 0})

    def test_bad_precol5(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.node_precoloring,
                      graph, precol={0: 0, 1: 3})


class TestEdgePrecolorings:
    def test_many(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            # color all nodes, then uncolor some to get a valid precoloring
            precol = gcol.edge_coloring(G, strategy="random")
            colSize = defaultdict(int)
            for e in precol:
                colSize[precol[e]] += 1
            E = list(precol)
            for i in range(len(E)):
                e = E[i]
                if i % 2 == 0 and colSize[precol[e]] > 1:
                    colSize[precol[e]] -= 1
                    del precol[e]
            for strategy in GREEDY_METHODS:
                for opt_alg in OPT_ALGS:
                    for it_limit in IT_LIMITS:
                        c = gcol.edge_precoloring(
                            G,
                            precol,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit,
                        )
                        assert verify_edge_coloring(G, c, precol)
                        c = gcol.edge_precolouring(
                            G,
                            precol,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit,
                        )
                        assert verify_edge_coloring(G, c, precol)
                        c = gcol.edge_precoloring(
                            G,
                            None,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit,
                        )
                        assert verify_edge_coloring(G, c)
                        c = gcol.edge_precoloring(
                            G,
                            {},
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit
                        )
                        assert verify_edge_coloring(G, c, {})

    def test_bad_precol1(self):
        graph = three_node_clique()
        pytest.raises(TypeError, gcol.edge_precoloring,
                      graph, precol="this is not a dict")

    def test_bad_precol2(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.edge_precoloring,
                      graph, precol={(1, 4): 1})

    def test_bad_precol3(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.edge_precoloring,
                      graph, precol={(0, 1): "color 5"})

    def test_bad_precol4(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.edge_precoloring,
                      graph, precol={(0, 1): 0, (1, 2): 0})

    def test_bad_precol5(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.edge_precoloring,
                      graph, precol={(0, 1): 0, (1, 2): 3})


class TestKempeChain:
    def test_many(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            c = gcol.node_coloring(G, strategy="random")
            k = get_num_cols(c)
            for u in G:
                for j in range(k):
                    gcol.kempe_chain(G, c, u, j)

    def test_clashing_col(self):
        graph = nx.erdos_renyi_graph(10, 1.0)
        c = {v: 0 for v in graph}
        pytest.raises(ValueError, gcol.kempe_chain, graph, c, 0, 0)

    def test_bad_s(self):
        graph = nx.erdos_renyi_graph(10, 1.0)
        c = gcol.node_coloring(graph, strategy="random")
        pytest.raises(ValueError, gcol.kempe_chain, graph, c, 999, 0)


class TestKColourings:
    def test_many(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            for opt_alg in OPT_ALGS:
                for it_limit in IT_LIMITS:
                    c = gcol.node_k_coloring(
                        G, len(G), opt_alg=opt_alg, it_limit=it_limit
                    )
                    assert verify_node_coloring(G, c)
                    c = gcol.node_k_colouring(
                        G, len(G), opt_alg=opt_alg, it_limit=it_limit
                    )
                    assert verify_node_coloring(G, c)
                    c = gcol.edge_k_coloring(
                        G,
                        get_max_degree(G) + 1,
                        opt_alg=opt_alg,
                        it_limit=it_limit
                    )
                    assert verify_edge_coloring(G, c)
                    c = gcol.edge_k_colouring(
                        G,
                        get_max_degree(G) + 1,
                        opt_alg=opt_alg,
                        it_limit=it_limit
                    )
                    assert verify_edge_coloring(G, c)
                    c = gcol.equitable_node_k_coloring(
                        G,
                        len(G),
                        weight=None,
                        opt_alg=opt_alg,
                        it_limit=it_limit
                    )
                    assert verify_node_coloring(G, c)
                    c = gcol.equitable_edge_k_coloring(
                        G,
                        get_max_degree(G) + 1,
                        weight=None,
                        opt_alg=opt_alg,
                        it_limit=it_limit,
                    )
                    assert verify_edge_coloring(G, c)

    def test_bad_k_1(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.node_k_coloring, graph, 2)
        pytest.raises(ValueError, gcol.edge_k_coloring, graph, 2)
        pytest.raises(ValueError, gcol.equitable_node_k_coloring, graph, 2)
        pytest.raises(ValueError, gcol.equitable_edge_k_coloring, graph, 2)

    def test_bad_k_2(self):
        graph = three_node_clique()
        pytest.raises(ValueError, gcol.node_k_coloring, graph, -2)
        pytest.raises(ValueError, gcol.edge_k_coloring, graph, -2)
        pytest.raises(ValueError, gcol.equitable_node_k_coloring, graph, -2)
        pytest.raises(ValueError, gcol.equitable_edge_k_coloring, graph, -2)

    def test_bad_node_weights(self):
        graph = nx.Graph()
        graph.add_nodes_from([0, 1, 2], weight=-5)
        graph.add_edges_from([(0, 1), (1, 2), (2, 0)])
        pytest.raises(
            ValueError,
            gcol.equitable_node_k_coloring, graph, 3, weight="weight"
        )

    def test_bad_edge_weights(self):
        graph = nx.Graph()
        graph.add_nodes_from([0, 1, 2])
        graph.add_edges_from([(0, 1), (1, 2), (2, 0)], weight=-5)
        pytest.raises(
            ValueError,
            gcol.equitable_edge_k_coloring, graph, 3, weight="weight"
        )


class TestMaxIS:
    def test_many(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            for it_limit in IT_LIMITS:
                S = gcol.max_independent_set(G, weight=None, it_limit=it_limit)
                assert verify_independent_set(G, S)


class TestMinCostKColoring:
    def test_many(self):
        for graph_func in TEST_CASES:
            G = graph_func()
            for u in G:
                for v in G[u]:
                    G[u][v]["weight"] = 3
            for it_limit in IT_LIMITS:
                for k in range(1, len(G) + 1):
                    gcol.min_cost_k_coloring(
                        G,
                        k,
                        weight=None,
                        weights_at="nodes",
                        it_limit=it_limit
                    )
                    gcol.min_cost_k_colouring(
                        G,
                        k,
                        weight=None,
                        weights_at="nodes",
                        it_limit=it_limit
                    )
                for k in range(2, len(G) + 1):
                    gcol.min_cost_k_coloring(
                        G,
                        k,
                        weight=None,
                        weights_at="edges",
                        it_limit=it_limit
                    )
                    gcol.min_cost_k_colouring(
                        G,
                        k,
                        weight=None,
                        weights_at="edges",
                        it_limit=it_limit
                    )
                for k in range(2, len(G) + 1):
                    gcol.min_cost_k_coloring(
                        G,
                        k,
                        weight="weight",
                        weights_at="edges",
                        it_limit=it_limit
                    )
                    gcol.min_cost_k_colouring(
                        G,
                        k,
                        weight="weight",
                        weights_at="edges",
                        it_limit=it_limit
                    )

    def test_bad_node_weights(self):
        graph = nx.Graph()
        graph.add_nodes_from([0, 1, 2], weight=-5)
        graph.add_edges_from([(0, 1), (1, 2), (2, 0)])
        pytest.raises(
            ValueError,
            gcol.min_cost_k_coloring,
            graph,
            3,
            weight="weight",
            weights_at="nodes",
        )

    def test_bad_edge_weight(self):
        graph = nx.Graph()
        graph.add_nodes_from([0, 1, 2])
        graph.add_edges_from([(0, 1), (1, 2), (2, 0)], weight=-5)
        pytest.raises(
            ValueError,
            gcol.min_cost_k_coloring,
            graph,
            3,
            weight="weight",
            weights_at="edges",
        )

    def test_bad_node_weight_labels(self):
        graph = nx.Graph()
        graph.add_nodes_from([0, 1, 2], weight=5)
        graph.add_edges_from([(0, 1), (1, 2), (2, 0)])
        pytest.raises(
            ValueError,
            gcol.min_cost_k_coloring,
            graph,
            3,
            weight="wait",
            weights_at="nodes",
        )

    def test_bad_edge_weight_labels(self):
        graph = nx.Graph()
        graph.add_nodes_from([0, 1, 2])
        graph.add_edges_from([(0, 1), (1, 2), (2, 0)], weight=5)
        pytest.raises(
            ValueError,
            gcol.min_cost_k_coloring,
            graph,
            3,
            weight="wait",
            weights_at="edges",
        )


class TestFaceColorings:

    def test_many(self):
        for graph_func in FACE_TEST_CASES:
            G = graph_func()
            pos = nx.planar_layout(G)
            for strategy in GREEDY_METHODS:
                for opt_alg in OPT_ALGS:
                    for it_limit in IT_LIMITS:
                        gcol.face_coloring(
                            G,
                            pos,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit
                        )
                        gcol.face_colouring(
                            G,
                            pos,
                            strategy=strategy,
                            opt_alg=opt_alg,
                            it_limit=it_limit
                        )

    def test_bad_strategy(self):
        graph = singleton()
        pos = {0: (0, 0)}
        pytest.raises(
            ValueError,
            gcol.face_coloring,
            graph,
            pos,
            strategy="this is an invalid strategy",
        )

    def test_bad_opt_alg(self):
        graph = singleton()
        pos = {0: (0, 0)}
        pytest.raises(
            ValueError,
            gcol.face_coloring,
            graph,
            pos,
            opt_alg="this is an invalid optimisation algorithm",
        )

    def test_bad_its_parameter(self):
        graph = singleton()
        pos = {0: (0, 0)}
        pytest.raises(
            ValueError,
            gcol.face_coloring,
            graph,
            pos,
            it_limit="this is not an integer",
        )

    def test_negative_tabu_parameter(self):
        graph = singleton()
        pos = {0: (0, 0)}
        pytest.raises(ValueError, gcol.face_coloring, graph, pos, it_limit=-1)

    def test_directed_graph(self):
        graph = nx.erdos_renyi_graph(3, 1.0, directed=True)
        pos = {0: (0, 0), 1: (1, 1), 3: (1, 0)}
        pytest.raises(NotImplementedError, gcol.face_coloring, graph, pos)

    def test_multigraph(self):
        graph = nx.MultiGraph()
        graph.add_edges_from(
            [(0, 1), (0, 1)])
        pos = {0: (0, 0), 1: (1, 1)}
        pytest.raises(NotImplementedError, gcol.face_coloring, graph, pos)

    def test_invalid_pos(self):
        graph = nx.complete_graph(3)
        pos = "I am not a dict"
        pytest.raises(TypeError, gcol.face_coloring, graph, pos)

    def test_missing_pos(self):
        graph = nx.complete_graph(3)
        pos = {0: (0, 0), 1: (1, 0)}
        pytest.raises(ValueError, gcol.face_coloring, graph, pos)

    def test_duplicate_pos(self):
        graph = nx.complete_graph(3)
        pos = {0: (0, 0), 1: (1, 0), 2: (1, 0)}
        pytest.raises(ValueError, gcol.face_coloring, graph, pos)

    def test_not_planar(self):
        graph = nx.complete_graph(4)
        pos = {0: (0, 0), 1: (1, 0), 2: (0, 1), 3: (1, 1)}
        pytest.raises(ValueError, gcol.face_coloring, graph, pos)

    def test_has_bridges(self):
        graph = nx.path_graph(3)
        pos = {0: (0, 0), 1: (1, 0), 2: (0, 1)}
        pytest.raises(ValueError, gcol.face_coloring, graph, pos)

    def test_not_embedding(self):
        graph = nx.cycle_graph(4)
        pos = {0: (0, 0), 1: (1, 0), 2: (0, 1), 3: (1, 1)}
        pytest.raises(ValueError, gcol.face_coloring, graph, pos)

    def test_not_embedding2(self):
        graph = nx.cycle_graph(4)
        pos = {0: (0, 0), 1: (1, 0), 2: (1, 1), 3: (2, 0)}
        pytest.raises(ValueError, gcol.face_coloring, graph, pos)

    def test_precol_invalid_arg(self):
        graph = nx.cycle_graph(4)
        graph.add_edge(0, 2)
        pos = {0: (0, 0), 1: (1, 0), 2: (1, 1), 3: (0, 1)}
        precol = "I am not a dict"
        pytest.raises(TypeError, gcol.face_precoloring, graph, pos, precol)

    def test_precol_invalid_edge(self):
        graph = nx.cycle_graph(4)
        graph.add_edge(0, 2)
        pos = {0: (0, 0), 1: (1, 0), 2: (1, 1), 3: (0, 1)}
        precol = {(1, 3): 1}
        pytest.raises(ValueError, gcol.face_precoloring, graph, pos, precol)

    def test_precol_invalid_col(self):
        graph = nx.cycle_graph(4)
        graph.add_edge(0, 2)
        pos = {0: (0, 0), 1: (1, 0), 2: (1, 1), 3: (0, 1)}
        precol = {(1, 2): 0.5}
        pytest.raises(ValueError, gcol.face_precoloring, graph, pos, precol)

    def test_precol_invalid_col2(self):
        graph = nx.cycle_graph(4)
        graph.add_edge(0, 2)
        pos = {0: (0, 0), 1: (1, 0), 2: (1, 1), 3: (0, 1)}
        precol = {(1, 2): 1}
        pytest.raises(ValueError, gcol.face_precoloring, graph, pos, precol)

    def test_precol_invalid_col3(self):
        graph = nx.cycle_graph(4)
        graph.add_edge(0, 2)
        pos = {0: (0, 0), 1: (1, 0), 2: (1, 1), 3: (0, 1)}
        precol = {(1, 0): 0, (2, 1): 1}
        pytest.raises(ValueError, gcol.face_precoloring, graph, pos, precol)


# Utility functions
def verify_node_coloring(G, c, precol=None):
    if len(c) != 0:
        k = max(c.values()) + 1
        used = set()
        for v in c:
            assert (
                c[v] >= 0 and c[v] < k
            ), "Node assigned to a color not in {0,1,...,k-1}"
            used.add(c[v])
        assert len(used) == k, "Unused colors in {0,1,...,k-1}"
        for u in G:
            for v in G[u]:
                assert c[u] != c[v], "Adjacent nodes have the same color"
        if precol is not None:
            for v in precol:
                assert (
                    c[v] == precol[v]
                ), "Error a precolored node is not colored correctly"
    gcol.partition(c)
    gcol.get_node_colors(G, c)
    return True


def verify_edge_coloring(G, c, precol=None):
    if len(c) != 0:
        k = max(c.values()) + 1
        used = set()
        for e in c:
            assert (
                c[e] >= 0 and c[e] < k
            ), "Edge assigned to a color not in {0,1,...,k-1}"
            used.add(c[e])
        assert len(used) == k, "Unused colors in {0,1,...,k-1}"
        for e1 in G.edges:
            for e2 in G.edges:
                if e1 != e2:
                    if (
                        e1[0] == e2[0]
                        or e1[0] == e2[1]
                        or e1[1] == e2[0]
                        or e1[1] == e2[1]
                    ):
                        assert c[e1] != c[e2], "Adjacent edges are same color"
        if precol is not None:
            for u, v in precol:
                assert (
                    c[u, v] == precol[u, v]
                ), "Error a precolored edge is not colored correctly"
    gcol.partition(c)
    gcol.get_edge_colors(G, c)
    return True


def verify_independent_set(G, S):
    for u in G:
        for v in G[u]:
            assert not (
                u in S and v in S
            ), "Adjacent nodes present in the independent set S"
    gcol.get_set_colors(G, S)
    return True


def get_max_degree(G):
    # Get the maximum degree in G
    return max(G.degree(node) for node in G.nodes) if len(G.nodes) > 0 else 0


def get_num_cols(c):
    if len(c) == 0:
        return 0
    return max(c.values()) + 1


# Different graphs to test
def null_graph():
    return nx.Graph()


def singleton():
    graph = nx.Graph()
    graph.add_node(0)
    return graph


def dyad():
    graph = nx.Graph()
    graph.add_edge(0, 1)
    return graph


def three_node_clique():
    graph = nx.Graph()
    graph.add_edges_from([(0, 1), (1, 2), (2, 0)])
    return graph


def disconnected():
    graph = nx.Graph()
    graph.add_edges_from([(1, 2), (2, 3), (4, 5), (5, 6)])
    return graph


def mixed_names():
    graph = nx.Graph()
    graph.add_nodes_from([0, 1, "node-A", ("A", 3)])
    graph.add_edges_from([(0, 1), (1, "node-A"),
                          ("node-A", ("A", 3)), (0, ("A", 3))])
    return graph


def empty():
    return nx.erdos_renyi_graph(10, 0.0, 1)


def sparse():
    return nx.erdos_renyi_graph(10, 0.1, 1)


def medium():
    return nx.erdos_renyi_graph(10, 0.5, 1)


def dense():
    return nx.erdos_renyi_graph(10, 0.9, 1)


def complete():
    return nx.erdos_renyi_graph(10, 1.0, 1)


def complete_bipartite():
    graph = nx.Graph()
    for u in range(0, 10, 2):
        for v in range(1, 11, 2):
            graph.add_edge(u, v)
    return graph


def dodec():
    return nx.dodecahedral_graph()


def grid():
    return nx.grid_2d_graph(3, 4)


# --------------------------------------------------------------------------
# Test graphs
TEST_CASES = [
    null_graph,
    singleton,
    dyad,
    disconnected,
    three_node_clique,
    mixed_names,
    empty,
    sparse,
    medium,
    dense,
    complete,
    complete_bipartite,
    dodec,
    grid,
]

FACE_TEST_CASES = [
    null_graph,
    singleton,
    three_node_clique,
    mixed_names,
    empty,
    dodec,
    grid,
]



"""Node coloring functions for the gcol library."""
import random
import itertools
import networkx as nx
from queue import PriorityQueue
from collections import defaultdict


def _greedy(G, V):
    # Greedy algorithm for graph coloring. This considers nodes of G in the
    # order given in V
    c = {}
    for u in V:
        adjcols = {c[v] for v in G[u] if v in c}
        for j in itertools.count():
            if j not in adjcols:
                break
        c[u] = j
    return c


def _dsatur(G, c=None):
    # Dsatur algorithm for graph coloring. First initialise the data
    # structures. These are: the colors of each node c[v]; the degree d[v] of
    # each uncolored node in the graph induced by uncolored nodes; the set of
    # colors adjacent to each uncolored node (initially empty sets); and a
    # priority queue q. In q, each element has 4 values for the node v. The
    # first two are the the saturation degree of v, d[v] (as a tie breaker).
    # The third value is a counter, which just stops comparisons being made
    # with the final values, which might be of different types.
    d, adjcols, q = {}, {}, PriorityQueue()
    counter = itertools.count()
    for u in G.nodes:
        d[u] = G.degree(u)
        adjcols[u] = set()
        q.put((0, d[u] * (-1), next(counter), u))
    # If any nodes are already colored in c, update the data structures
    # accordingly
    if c is not None:
        if not isinstance(c, dict):
            raise TypeError(
                "Error, c should be a dict that assigns a subset of nodes ",
                "to colors"
            )
        for u in c:
            for v in G[u]:
                if v not in c:
                    adjcols[v].add(c[u])
                    d[v] -= 1
                    q.put((len(adjcols[v]) * (-1), d[v]
                          * (-1), next(counter), v))
                elif c[u] == c[v]:
                    raise ValueError(
                        "Error, clashing nodes defined in supplied coloring"
                    )
    else:
        c = {}
        # Color all remaining nodes
    while len(c) < len(G):
        # Get the uncolored node u with max saturation degree, breaking ties
        # using the highest value for d. Remove u from q.
        _, _, _, u = q.get()
        if u not in c:
            # Get lowest color label i for uncolored node u
            for i in itertools.count():
                if i not in adjcols[u]:
                    break
            c[u] = i
            # Update the data structures
            for v in G[u]:
                if v not in c:
                    adjcols[v].add(i)
                    d[v] -= 1
                    q.put((len(adjcols[v]) * (-1), d[v]
                          * (-1), next(counter), v))
    return c


def _rlf(G):
    def update_rlf(u):
        # Remove u from X (it has been colored) and move all uncolored
        # neighbors of u from X to Y
        X.remove(u)
        for v in G[u]:
            if v not in c:
                X.discard(v)
                Y.add(v)
        # Recalculate the contets of NInX and NInY. First calculate a set D2
        # of all uncolored nodes within distance two of u.
        D2 = set()
        for v in G[u]:
            if v not in c:
                D2.add(v)
                for w in G[v]:
                    if w not in c:
                        D2.add(w)
        # For each node v in D2, recalculate the number of (uncolored)
        # neighbors in X and Y
        for v in D2:
            NInX[v] = 0
            NInY[v] = 0
            for w in G[v]:
                if w not in c:
                    if w in X:
                        NInX[v] += 1
                    elif w in Y:
                        NInY[v] += 1

    # RLF algorithm for graph coloring. Here, X is the set of uncolored nodes
    # not adjacent to any nodes colored with color i, and Y is the set of
    # uncolored nodes that are adjcent to nodes colored with i.
    c, Y, n, i = {}, set(), len(G), 0
    X = set(G.nodes())
    while X:
        # Construct color class i. First, for each nodes u in X, calculate the
        # number of neighbors it has in X and Y
        NInX, NInY = {u: 0 for u in X}, {u: 0 for u in X}
        for u in X:
            for v in G[u]:
                if v in X:
                    NInX[u] += 1
        # Identify and colur the uncolored node u in X that has the most
        # neighbors in X
        maxVal = -1
        for v in X:
            if NInX[v] > maxVal:
                maxVal, u = NInX[v], v
        c[u] = i
        update_rlf(u)
        while X:
            # Identify and color the node u in X that has the largest number
            # of neighbors in Y. Break ties according to the min neighbors in X
            mxVal, mnVal = -1, n
            for v in X:
                if NInY[v] > mxVal or (NInY[v] == mxVal and NInX[v] < mnVal):
                    mxVal, mnVal, u = NInY[v], NInX[v], v
            c[u] = i
            update_rlf(u)
        # Have finished constructing color class i
        X, Y = Y, X
        i += 1
    return c


def _backtrackcol(G, targetcols, verbose):
    def is_feasible(u, i):
        # Returns true iff node u can be feasibly assigned to color i in c
        for v in G[u]:
            if c.get(v) == i:
                return False
        return True

    def color(uPos):
        # Recursive function used for backtracking. Attempts to color node at
        # position uPos in V
        its[0] += 1
        if len(colsize) > numcols[0]:
            # Current (partial) solution is using too many colors, so backtrack
            return False
        if uPos == len(G):
            # At a leaf node in search tree. A new best solution has been
            # found.
            bestc.clear()
            for v in c:
                bestc[v] = c[v]
            if verbose > 0:
                print("    Found solution with", len(colsize),
                      "colors. Total backtracking iterations =", its[0])
            if len(colsize) == targetcols:
                # Optimum solution has been constructed or target reached
                return True
            else:
                # Reduce number of available colors and continue
                numcols[0] = len(colsize) - 1
                return False
        u = V[uPos]
        for i in range(numcols[0]):
            if i < numcols[0] and is_feasible(u, i):
                c[u] = i
                colsize[i] += 1
                if color(uPos + 1):
                    return True
                colsize[c[u]] -= 1
                if colsize[c[u]] == 0:
                    del colsize[c[u]]
                del c[u]
        return False

    # Find a large clique C in G
    C = list(nx.approximation.max_clique(G))
    targetcols = max(targetcols, len(C))
    if verbose > 0:
        print("Running backtracking algorithm:")
    # Generate an initial solution. Do this by assigning the nodes in C
    # to different colors, then get a starting number of colors (numcols) using
    # dsatur. V holds the order in which the vetices were colored
    bestc = {C[i]: i for i in range(len(C))}
    bestc = _dsatur(G, bestc)
    numcols = [max(bestc.values()) + 1]
    if verbose > 0:
        print("    Found solution with",
              numcols[0], "colors. Total backtracking iterations = 0")
    numcols[0] -= 1
    V = list(bestc)
    # Now assign the nodes in C to c and run the backtracking algorithm
    # from the next node in V. Here, bestc holds the best solution seen so far
    # and colsize holds the size of all nonempty color classes in c.
    # len(colsize) therefore gives the number of colors (cost) being used by
    # the (sub-)solution c
    c, colsize, its = {}, defaultdict(int), [0]
    for i in range(len(C)):
        c[C[i]] = i
        colsize[i] += 1
    color(len(C))
    if verbose > 0:
        print("Ending backtracking at iteration",
              its[0], "- optimal solution achieved.")
    return bestc


def _partialcol(G, k, c, W, it_limit, verbose):
    def domovepartialcol(v, j):
        # Used by partialcol to move node v to color j and update relevant
        # data structures
        c[v] = j
        U.remove(v)
        for u in G[v]:
            C[u, j] += W[v]
            if c[u] == j:
                T[u, j] = its + t
                U.add(u)
                c[u] = -1
                for w in G[u]:
                    C[w, j] -= W[u]

    # Use the current solution c to populate the data structures. C[v,j] gives
    # the total weight of the neighbors of v in color j, T is the tabu list,
    # and U is the set of clashing nodes
    assert k >= 1, "Error, partialcol only works with at least k = 1 color"
    C, T, U, its = {}, {}, set(), 0
    for v in G:
        assert (
            isinstance(c[v], int) and c[v] >= -1 and c[v] < k
        ), ("Error, the coloring defined by c must allocate each node a ",
            "value from the set {-1,0,...,k-1}, where -1 signifies that ",
            "a node is uncolored")
        for j in range(k):
            C[v, j] = 0
            T[v, j] = 0
    for v in G:
        if c[v] == -1:
            U.add(v)
        for u in G[v]:
            if c[u] != -1:
                C[v, c[u]] += W[u]
    currentcost = sum(W[u] for u in U)
    bestcost, bestsol, t = float("inf"), {}, 1
    if verbose > 0:
        print("    Running PartialCol algorithm using", k, "colors")
    while True:
        # Keep track of best solution and halt when appropriate
        if currentcost < bestcost:
            if verbose > 0:
                print("        Solution with", k, "colors and cost",
                      currentcost, "found by PartialCol at iteration", its)
            bestcost = currentcost
            bestsol = dict(c)
        if bestcost <= 0 or its >= it_limit:
            break
        # Evaluate all neighbors of current solution c
        its += 1
        vbest, jbest, bestval, numbestval = -1, -1, float("inf"), 0
        for v in U:
            for j in range(k):
                neighborcost = currentcost + C[v, j] - W[v]
                if neighborcost <= bestval:
                    if neighborcost < bestval:
                        numbestval = 0
                    # Consider the move if it is not tabu or leads to a new
                    # best solution
                    if T[v, j] < its or neighborcost < bestcost:
                        if random.randint(0, numbestval) == 0:
                            vbest, jbest, bestval = v, j, neighborcost
                        numbestval += 1
        # Do the chosen move. If no move was chosen (all moves are tabu),
        # choose a random move
        if vbest == -1:
            vbest = random.choice(tuple(U))
            jbest = random.randint(0, k - 1)
            bestval = currentcost + C[vbest, jbest] - W[vbest]
        # Apply the move, update T, and determine the next tabu tenure t
        domovepartialcol(vbest, jbest)
        currentcost = bestval
        t = int(0.6 * len(U)) + random.randint(0, 9)
    if verbose > 0:
        print("    Ending PartialCol")
    return bestcost, bestsol, its


def _tabucol(G, k, c, W, it_limit, verbose):
    def domovetabucol(v, j):
        # Used by tabucol to move node v to a new color j and update relevant
        # data structures
        i = c[v]
        c[v] = j
        if C[v, i] > 0 and C[v, j] == 0:
            U.remove(v)
        elif C[v, i] == 0 and C[v, j] > 0:
            U.add(v)
        for u in G[v]:
            C[u, i] -= W[v, u]
            if C[u, i] == 0 and c[u] == i:
                U.remove(u)
            C[u, j] += W[v, u]
            if C[u, j] > 0 and c[u] == j:
                U.add(u)
        T[v, i] = its + t

    assert k >= 2, "Error, tabucol only works with at least k = 2 colors"
    # Use the current solution c to populate the data structures. C[v,j] gives
    # the number of neighbors of v in color j, T is the tabu list, and U is the
    # set of clashing nodes
    C, T, U, its, currentcost = {}, {}, set(), 0, 0
    for v in G:
        assert isinstance(c[v], int) and c[v] >= 0 and c[v] < k, (
            "Error, the coloring defined by c must allocate each node a ",
            "value from the set {0,...,k-1}"
            + str(v)
            + " "
            + str(c[v])
        )
        for j in range(k):
            C[v, j] = 0
            T[v, j] = 0
    for v in G:
        for u in G[v]:
            C[v, c[u]] += W[v, u]
    for v in G:
        if C[v, c[v]] > 0:
            currentcost += C[v, c[v]]
            U.add(v)
    currentcost //= 2
    bestcost, bestsol, t = float("inf"), {}, 1
    if verbose > 0:
        print("    Running TabuCol algorithm using", k, "colors")
    while True:
        # Keep track of best solution and halt when appropriate
        if currentcost < bestcost:
            if verbose > 0:
                print("        Solution with", k, "colors and cost",
                      currentcost, "found by TabuCol at iteration", its)
            bestcost = currentcost
            bestsol = dict(c)
        if bestcost <= 0 or its >= it_limit:
            break
        # Evaluate all neighbors of current solution
        its += 1
        vbest, jbest, bestval, numbestval = -1, -1, float("inf"), 0
        for v in U:
            for j in range(k):
                if j != c[v]:
                    neighborcost = currentcost + C[v, j] - C[v, c[v]]
                    if neighborcost <= bestval:
                        if neighborcost < bestval:
                            numbestval = 0
                        # Consider the move if it is not tabu or leads to a new
                        # global best
                        if T[v, j] < its or neighborcost < bestcost:
                            if random.randint(0, numbestval) == 0:
                                vbest, jbest, bestval = v, j, neighborcost
                            numbestval += 1
        # Do the chosen move. If no move was chosen (all moves are tabu),
        # choose a random move
        if vbest == -1:
            vbest = random.choice(tuple(c))
            while True:
                jbest = random.randint(0, k - 1)
                if jbest != c[vbest]:
                    break
            bestval = currentcost + C[vbest, jbest] - C[vbest, c[vbest]]
        domovetabucol(vbest, jbest)
        currentcost = bestval
        t = int(0.6 * len(U)) + random.randint(0, 9)
    if verbose > 0:
        print("    Ending TabuCol")
    return bestcost, bestsol, its


def _removeColor(c, j, alg):
    maxcol = max(c.values())
    # Uncolor nodes assigned to color j while maintaining use of colors
    # 0,1,...,maxcol-1
    for v in c:
        if c[v] == j:
            c[v] = -1
        elif c[v] == maxcol:
            c[v] = j
    # If tabucol is being used, assign uncolored nodes to random colors
    if alg == 2:
        for v in c:
            if c[v] == -1:
                c[v] = random.randint(0, maxcol - 1)


def _reducecolors(G, c, target, W, opt_alg, it_limit, verbose):
    # Uses specified optimisation algorithm to try to reduce the number of
    # colors in c to the target value. The observed proper solution with the
    # fewest colors is returned (which may be using more colors than the
    # target)
    k = max(c.values()) + 1
    if opt_alg == 1:
        return _backtrackcol(G, target, verbose)
    bestc, totalits = dict(c), 0
    if verbose > 0:
        print("Running local search algorithm:")
        print("    Found solution with", k,
              "colors. Total local search iterations = 0 /", it_limit)
    while k > target and totalits < it_limit:
        k -= 1
        j = random.randint(0, k - 1)
        _removeColor(c, j, opt_alg)
        if opt_alg == 2:
            cost, c, its = _tabucol(
                G, k, c, W, it_limit - totalits, verbose - 1)
        else:
            cost, c, its = _partialcol(
                G, k, c, W, it_limit - totalits, verbose - 1)
        totalits += its
        if cost == 0:
            bestc = dict(c)
            if verbose > 0:
                print("    Found solution with", k,
                      "colors. Total local search iterations =", totalits,
                      "/", it_limit)
    if verbose > 0:
        if totalits >= it_limit:
            print("Ending local search. Iteration limit of",
                  it_limit, "has been reached.")
        else:
            print("Ending local search at iteration", totalits,
                  "- optimal solution achieved.")
    return bestc
